import re
import quetzpy.modules.data_cleaning as data_cleaning

def hashtag_extractor(tweet_text: str) -> list:
    """
    Extract hashtags from a tweet\n
    Returns a list of hashtags
    """

    try:
        # hashtags = pattern.match(tweet_text)
        hashtags = re.findall(r"#(\w+)", tweet_text)

    except:
        hashtags = ['']

    return hashtags

def mention_extractor(tweet_text: str) -> list:
    """
    Extract mentions from a tweet\n
    Returns a list of mentions
    """
    try:
        # hashtags = pattern.match(tweet_text)
        mentions = re.findall(r"@(\w+)", tweet_text)

    except:
        mentions = ['']

    return mentions

def sub_topics_extractor(tweets_text_list : list, lang_ = 'fr_core_news_sm', max_most_common = 5) -> any:
    """
    Extract sub_topicts\n
    take all the tweets' text as parameter\n
    return a dict of sub_topics
    """

    # Load the correct language for spacy
    import spacy
    from spacy.lang.fr.stop_words import STOP_WORDS
    from string import punctuation
    from collections import Counter

    # IMPORTANT : You must download the language package from spacy
    nlp = spacy.load(lang_)

    # Concatenate the list of texts
    big_text = data_cleaning.get_big_text(tweets_text_list)
    big_text = data_cleaning.emojies_remover(big_text)

    doc = nlp(big_text)

    keywords = []
    stopwords = list(STOP_WORDS)
    pos_tag = ['PROPN', 'ADJ', 'NOUN', 'VERB']
    for token in doc:
        if(token.text in stopwords or token.text in punctuation):
            continue
        if token.pos_ in pos_tag:
            keywords.append(token.text)
    freq_words = Counter(keywords)

    return freq_words.most_common(max_most_common)

def emojies_extractor(tweets_text_list : list, lang_ = 'fr_core_news_sm', max_most_common = 5) -> any:
    """
    Extract occurrence of emojies\n
    take all the tweets' text as parameter\n
    return a dict of most used emojies
    """
    import spacy
    from spacymoji import Emoji
    from collections import Counter

    nlp = spacy.load(lang_)
    nlp.add_pipe("emoji", first = True)

    # Concatenate the list of texts
    big_text = data_cleaning.get_big_text(tweets_text_list)

    doc = nlp(big_text)
    emojies = [emoji[0] for emoji in doc._.emoji]
    freq_emojies = Counter(emojies)

    return freq_emojies.most_common(max_most_common)


def track_words(self, Params_manager, topic : str) -> dict:
    """
    View how many times a word is detected in the tweets\n
    Take a list of words to track\n
    Returns a dict of the tracked words and the associated occurrence
    """

    # Initialize the dict of {words: occurrence}
    dict_tracker = dict()
    for to_track_word in Params_manager.Extrac_track_words:
        result = sum([1 for key, val in self.data_topics[topic]["tweets"].items() if to_track_word.lower() in val["text_cleaned"]])
        dict_tracker[to_track_word] = result

    return dict_tracker

def track_all():
    print("foo")
    return "foo"

def show_occurrence(self, Params_manager, topic) -> dict:
    """
    View the occurrence of either linked hashtags or mentions\n
    Take a parameter to know if it shall analyze hashtags or mentions\n
    Returns a dict of words and the associated occurrence
    """

    results = dict()
    
    for show_this_occurence in Params_manager.Extrac_show_occurrence:
        all_words = []
        for key, val in self.data_topics[topic]["tweets"].items():
            all_words.extend(val[show_this_occurence])
            
        all_words_uniques = set(all_words)

        result_for_this = {item:all_words.count(item) for item in all_words_uniques}
        result_for_this_sorted = dict(filter(lambda item_sort : item_sort[1] >= Params_manager.Extrac_show_occurrence_threshold, sorted(result_for_this.items(), key = lambda item : item[1], reverse = True)))
        results[show_this_occurence] = result_for_this_sorted

    return results


class Params_manager():

    def __init__(self, name_config_extractor = "/configuration_extractor.json", name_config_runner = "/configuration_runner.json", create_files = True) -> None:
        """
        Create and initialize `configuration_runner.json` and `extractor_configuration.json` used by the `Extractor` and the `Runner` objects\n
        :name_config_extractor: name for the extractor configuration file. Default = "/configuration_extractor.json"\n
        :name_config_runner: name for the runner configuration file. Default = "/configuration_runner.json"\n
        Invoking a Params_manager automatically creates the files
        """
        self.params_extractor = {
            "bearer token" : "MY TOKEN",
            "words to catch" : ["word 1", "word 2"],
            "topics" : ["topic 1", "topic 2"],
            "lang" : "en",
            "max results" : 10
        }

        self.params_runner = {
            "numbers of runs" : 1,
            "sleeping time (sec)" : 0
        }

        self.path = "./__config__"
        self.name_config_extractor = name_config_extractor
        self.name_config_runner = name_config_runner

        if create_files:
            self.create_files()

    def create_files(self) -> None:
        """
        Create the two files recquired by the `Extractor` class
        """

        import os
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        
        import json
        if not os.path.exists(self.path + self.name_config_extractor):        
            json.dump(self.params_extractor, open(self.path + self.name_config_extractor, "w"), indent = 4)

        if not os.path.exists(self.path + self.name_config_runner):
            json.dump(self.params_runner, open(self.path + self.name_config_runner, "w"), indent = 4)

        return

    def load_extrac_params(self) -> None:
        """
        Load parameters for `Extractor`\n
        `Extractor` parameters' names begin with `Extrac_`
        """

        import json
        path_to_config_extrac_file = self.path + self.name_config_extractor
        config_extractor = json.load(open(path_to_config_extrac_file))
        self.Extrac_bearer_token = config_extractor["bearer token"]
        self.Extrac_topics = config_extractor["topics"]
        self.Extrac_lang = config_extractor["lang"]
        self.Extrac_max_results = config_extractor["max results"]
        self.Extrac_track_words = config_extractor["stats"]["track words"]
        self.Extrac_track_words_threshold = config_extractor["stats threshold"]["track words"]
        self.Extrac_show_occurrence = config_extractor["stats"]["show occurrence"]
        self.Extrac_show_occurrence_threshold = config_extractor["stats threshold"]["show occurrence"]
        return

    def load_run_params(self) -> None:
        """
        Load parameters for `Runner`\n
        `Runner` parameters' names begin with `Run_`
        """
        import json
        path_to_runner_config_file = self.path + self.name_config_runner
        config_runner = json.load(open(path_to_runner_config_file))
        self.Run_nb_runs = config_runner["numbers of runs"]
        self.Run_sleep_time = config_runner["sleeping time (sec)"]

        return

    def load_params(self) -> None:
        self.load_extrac_params()
        self.load_run_params()



import re

def hashtags_remover(tweet_text: str) -> str:
    """
    Remove the hashtags from a tweet\n
    :tweet_text: text from a tweet\n
    :return: tweet_text cleaned from hashtags
    """
    pattern = re.compile(r"\s#+\S+")

    tweet_text_cleaned = pattern.sub("", tweet_text)

    return tweet_text_cleaned

def mentions_remover(tweet_text: str) -> str:
    """
    Remove the mentions from a tweet\n
    :tweet_text: text from a tweet\n
    :return: tweet_text cleaned from mentions
    """
    pattern = re.compile(r"\s@+\S+")

    tweet_text_cleaned = pattern.sub("", tweet_text)

    return tweet_text_cleaned

def websites_remover(tweet_text: str) -> str:
    """
    Remove the websites from a tweet\n
    :tweet_text: text from a tweet\n
    :return: tweet_text cleaned from websites
    """
    pattern = re.compile(r"\S+\\+\S+")
    tweet_text_cleaned_1 = pattern.sub("", tweet_text)

    pattern = re.compile(r"\s+http+\S+")

    tweet_text_cleaned_2 = pattern.sub("", tweet_text_cleaned_1)

    return tweet_text_cleaned_2

def remove_duplicates(self) -> str:
    """
    Remove duplicated tweets\n
    - Make an item_id from text and author_id\n
    - Drop duplicate based on item_id\n
    :return: a list of tweets cured from duplicates
    """
    items_id = []

    for item in self._data:
        length_text_quater = len(self._data[item]["text"])//4
        item_id = self._data[item]["text"][:length_text_quater] + str(self._data[item]["author_id"])
        items_id.append(item_id)
    
    set_items_id = set(items_id)

    if len(set_items_id) < len(items_id):
        temp = []
        new_data = {}
        for item in self._data:
            length_text_quater = len(self._data[item]["text"])//4
            if self._data[item]["text"][:length_text_quater] + str(self._data[item]["author_id"]) not in temp:
                temp.append(self._data[item]["text"][:length_text_quater] + str(self._data[item]["author_id"]))
                new_data.update(self._data[item])

        self._data = new_data
    
        return
    
    else:
        return self._data

def get_big_text(tweets_text_list : list) -> str:
    """
    Make a big text with all the tweets
    :tweets_text_list: list of tweets texts
    :return: a string that is the concatenation of tweets texts
    """

    big_text = ' '.join(tweets_text_list)
    big_text = big_text.replace('\n', ' ')
    big_text = re.sub(r'https?:\/\/.\S+', "", big_text)
    
    return big_text

def emojies_remover(big_text : str, lang_ = 'fr_core_news_sm'):
    """
    removes the emojies from the big text
    :big_text: a string that is the concatenation of tweets texts
    :lang_: language used to interprete the emojies
    :return: the big_text cured from emojies
    """

    import spacy
    from spacymoji import Emoji

    nlp = spacy.load(lang_)
    nlp.add_pipe("emoji", first = True)

    doc = nlp(big_text)
    emojies = set([emoji[0] for emoji in doc._.emoji])

    for emoji in emojies:
        big_text = big_text.replace(emoji, '')

    return big_text


from quetzpy.params_manager import Params_manager
import quetzpy.modules.twitter_connection as connect
import quetzpy.modules.data_cleaning as data_cleaning
import quetzpy.modules.data_mining as data_mining
import quetzpy.modules.data_stats as data_stats
import quetzpy.modules.decorators as decorators

class Extractor():
    def __init__(self, Params_manager : Params_manager):
        """
        Main class\n
        - Extract tweets based on a given topic\n
        - Can extract hashtags, mentions, most dicussed topics in the tweets
        - Can track information you give it
        """
        self.params_manager = Params_manager

        self.actions_states = {
            "clean tweets" : True,
            "tweet miner" : False,
            "tweets stats" : False
        }

        self.perform_actions = {
            "extract tweets" : self.extract_tweets,
            "clean tweets" : self.tweet_cleaner,
            "tweet miner" : self.tweet_miner,
            "tweets stats" : self.tweets_stats
        }

        self.tweets_stats_param = "all"

        return

    def activate_action(self, action : str):
        """
        Switch action state to True in the self.activate_actions dict\n
        List on actions to activate :\n
        * tweet miner
        * tweet stats
        """
        if action not in self.actions_states.keys():
            raise ValueError(f"Action given is wrong. It should be from this list {list(self.activate_actions.keys())[1::]}")
        
        # Switch action to True to be performed
        self.actions_states[action] = True

        return
    
    def extract_tweets(self):
        """
        Use tweeter_connection module to :\n
        * Connect to Tweeter REST API\n
        * Run the query\n
        * Split the dataset between the RT tweets and the non-RT ones
        """
        twitter_api = connect.connect_twitter(self.params_manager.Extrac_bearer_token)
        data_topics = dict()
        for topic in self.params_manager.Extrac_topics:
            data_topics[topic] = dict()
            data_topics[topic]["tweets"] = connect.query_tweeter(twitter_api = twitter_api, topic = topic, lang = self.params_manager.Extrac_lang, max_results = self.params_manager.Extrac_max_results)

        self.data_topics = data_topics
        return
    
    def tweet_cleaner(self):
        """
        Clean text from tweet
        """

        for topic in self.params_manager.Extrac_topics:
            for item in self.data_topics[topic]["tweets"]:
                self.data_topics[topic]["tweets"][item]['text_cleaned'] = data_cleaning.hashtags_remover(self.data_topics[topic]["tweets"][item]['text'].lower())
                self.data_topics[topic]["tweets"][item]['text_cleaned'] = data_cleaning.mentions_remover(self.data_topics[topic]["tweets"][item]['text_cleaned'])
                self.data_topics[topic]["tweets"][item]['text_cleaned'] = data_cleaning.websites_remover(self.data_topics[topic]["tweets"][item]['text_cleaned'])

        return

    def tweet_miner(self):
        """
        Extract data from tweet
        """
        for topic in self.params_manager.Extrac_topics:
            for item in self.data_topics[topic]["tweets"]:
                self.data_topics[topic]["tweets"][item]['hashtags'] = data_mining.hashtag_extractor(self.data_topics[topic]["tweets"][item]['text'].lower())
                self.data_topics[topic]["tweets"][item]['mentions'] = data_mining.mention_extractor(self.data_topics[topic]["tweets"][item]['text'].lower())

            self.data_topics[topic]["sub topics"] = data_mining.sub_topics_extractor([val['text_cleaned'] for val in self.data_topics[topic]["tweets"].values()])
            self.data_topics[topic]["emojies"] = data_mining.emojies_extractor([val['text_cleaned'] for val in self.data_topics[topic]["tweets"].values()])

        return

    def set_tweets_stats(self, new_param: str):
        """
        Set the tweets stats parameter to one of the following:\n
        * track words: Track only the words mentioned in the config file\n
        * show occurence:  Show the number of times hashtags or mention appears. Default: on hashtags\n
        * all: Run track word and show occurrence\n
        By default, the parameter is set to all.
        """
        if new_param not in ["all", "track words", "show occurrence"]:
            return "New parameter should be from this list:\nall, track words, show occurrence"
        self.tweets_stats_param = new_param

    def tweets_stats(self):
        """
        Compute statistics for the extracted tweets
        """
        stats_menu = {
            "track words": data_stats.track_words,
            "show occurrence": data_stats.show_occurrence,
            "all": data_stats.track_all
            }
        for topic in self.params_manager.Extrac_topics:
            self.data_topics[topic]["stats"] = stats_menu[self.tweets_stats_param](self, self.params_manager, topic)

        return



from quetzpy.params_manager import Params_manager
from quetzpy.extractor import Extractor

class Runner():
    def __init__(self, Params_manager : Params_manager, Extractor : Extractor) -> None:
        """
        Run the extraction according to the parameters
        """
        self.params_manager = Params_manager
        self.extractor = Extractor
        return

    def run_extractor_actions(self):
        """
        Run the Extractor's actions with state = True
        """

        extrac_actions_true = [action for action in self.extractor.actions_states.keys() if self.extractor.actions_states[action] == True]
        for action_to_perform in extrac_actions_true:
            self.extractor.perform_actions[action_to_perform]()

        return

    def write_output(self, directory = "./data"):
        """
        Write the output in the `./data` directory
        """
        import os
        if not os.path.exists(directory):
            os.mkdir(directory)

        from datetime import datetime
        now = datetime.now().strftime('%d%m%Y')

        for topic in self.extractor.params_manager.Extrac_topics:
            file_name = f"/data_{now}_{topic}.json"

            import json
            if not os.path.exists(directory + file_name):
                json.dump({"data" : []}, open(directory + file_name, "w"), indent = 4, ensure_ascii = False)

            with open(directory + file_name, "r+", encoding="utf-8") as f:
                f_data = json.load(f)
                f_data["data"].append(self.extractor.data_topics[topic])
                f.seek(0)
                json.dump(f_data, f, indent = 4, ensure_ascii=False)
    
        return

    def run_extraction(self):
        """
        Run the extraction according to the parameters
        """

        current_run = 0
        from time import sleep

        while current_run < self.params_manager.Run_nb_runs:
            
            self.extractor.extract_tweets()
            self.run_extractor_actions()
            self.write_output()

            current_run += 1
            if current_run == self.params_manager.Run_nb_runs:
                break

            else:
                sleep(self.params_manager.Run_sleep_time)


import tweepy as tw

def connect_twitter(bearer_token):
    """
    Connect to tweeter
    """
    twitter_api = tw.Client(bearer_token=bearer_token)
    
    return twitter_api

def query_tweeter(twitter_api, topic, lang = "", max_results = 100):
    """
    Extract 100 tweets on the topic\n
    """
    # Extract the list of tweets
    try :
        search = twitter_api.search_recent_tweets(query=f"{topic} lang:{lang} -is:retweet", 
            tweet_fields = ['author_id','created_at','text','source','lang'],
            user_fields = ['name','username','location','verified'],
            max_results = max_results
            )
    except:
        search = twitter_api.search_recent_tweets(query=f"{topic} -is:retweet", 
            tweet_fields = ['author_id','created_at','text','source','lang'],
            user_fields = ['name','username','location','verified'],
            max_results = max_results
            )

    _data = {"item_" + str(i): search.data[i].data for i in range(len(search.data))}
    
    return _data